#!/usr/bin/perl
use strict;
use bigint;

my $count = 0;
my $base = 64;
my $longsize = 16;
my $columnsize = 2;
my $dbtype = 1;
my $filename_in = "IPV6-COUNTRY.CSV";
my $filename_out = "IPV6-COUNTRY.BIN";
my %country;
my %db;

open IN, "<$filename_in" or die;
while (<IN>) {
	my @array = &splitcsv($_);
	$db{$array[0]}{"country"} = $array[4];
	$country{$array[4]}{"LONG"} = $array[5];
	$count++;
}
close IN;

$country{"-"}{"LONG"} = "-";
$count++;

my $addr = $base + $count * ( $longsize + 4 * ($columnsize - 1) );

foreach my $co (sort keys(%country)){
	$country{$co}{"ADDR"} = $addr;
	$addr = $addr + 1 + 2 + 1 + length($country{$co}{"LONG"});
}

my ($Second, $Minute, $Hour, $Day, $Month, $Year, $WeekDay, $DayOfYear, $IsDST) = localtime(time) ; 

open OUT, ">$filename_out" or die;
binmode(OUT);

print OUT pack("C", $dbtype);
print OUT pack("C", $columnsize);
print OUT pack("C", $Year - 100);
print OUT pack("C", $Month + 1);
print OUT pack("C", $Day);
print OUT pack("V", $count);
print OUT pack("V", 64+1);
print OUT pack("C", 1);

foreach my $i (14..63) {
	print OUT pack("C", 0);
}

foreach my $ip (sort { &bigintsort($a, $b); } keys(%db)) {
	print OUT &int2bytes($ip);
	print OUT pack("V", $country{$db{$ip}{"country"}}{"ADDR"});
}

print OUT &int2bytes(340282366920938463463374607431768211455);
print OUT pack("V", $country{"-"}{"ADDR"});

foreach my $co (sort keys(%country)) {
	print OUT pack("C", length($co));
	print OUT $co;
	if ($co eq "-") {
		print OUT " ";
	}
	print OUT pack("C", length($country{$co}{"LONG"}));
	print OUT $country{$co}{"LONG"};
}

close OUT;

sub bigintsort {
	my $inta = new Math::BigInt(shift(@_));
	my $intb = new Math::BigInt(shift(@_));
	return $inta <=> $intb;
}

sub int2hex {
	my $ip = new Math::BigInt(shift(@_));
	my $n = 0;
	my $binip = "";
	foreach my $i (1..128)	{
		($ip, $n) = $ip->bdiv(2);
		$binip = $n . $binip;
	}
	return join (':', unpack( 'H4H4H4H4H4H4H4H4', pack( 'B128', $binip )));
}

sub int2bytes {
	my $ip = new Math::BigInt(shift(@_));
	my $binip1_31 = 0;
	my $binip32_63 = 0;
	my $binip64_95 = 0;
	my $binip96_127 = 0;

	($ip, $binip1_31) = $ip->bdiv(4294967296);
	($ip, $binip32_63) = $ip->bdiv(4294967296);
	($ip, $binip64_95) = $ip->bdiv(4294967296);
	($ip, $binip96_127) = $ip->bdiv(4294967296);
	
	return pack("V", $binip1_31) . pack("V", $binip32_63) .pack("V", $binip64_95) .pack("V", $binip96_127);
}

sub bytes2int {
	my $binip = shift(@_);
	my @array = split(//, $binip);
	return 0 if ($#array != 15);
	my $ip96_127 = unpack("V", $array[0] . $array[1] . $array[2] . $array[3]);
	my $ip64_95 = unpack("V", $array[4] . $array[5] . $array[6] . $array[7]);
	my $ip32_63 = unpack("V", $array[8] . $array[9] . $array[10] . $array[11]);
	my $ip1_31 = unpack("V", $array[12] . $array[13] . $array[14] . $array[15]);

	return ($ip1_31 * 4294967296 * 4294967296 * 4294967296) + ($ip32_63 * 4294967296 * 4294967296) + ($ip64_95 * 4294967296) + $ip96_127;
}

sub hex2int {
	my $hexip = shift(@_);
	$hexip =~ s/://g;
	unless (length ($hexip) == 32) {
		return 0;
	};
	
	my $binip = unpack( 'B128', pack( 'H32', $hexip ));	
	my ($n, $dec) = (Math::BigInt->new(1), Math::BigInt->new(0));
	foreach (reverse (split '', $binip)) {
		$_ and $dec += $n;
		$n*=2;
	};	

	$dec=~s/^\+//;
	return $dec;
}

sub splitcsv() {
	my $line = shift (@_);
	my $sep = ',';
	return () unless $line;
	my @cells;
	$line =~ s/\r?\n$//;
	my $re = qr/(?:^|$sep)(?:\"([^\"]*)\"|([^$sep]*))/;
	while($line =~ /$re/g) {
		my $value = defined $1 ? $1 : $2;
		push @cells, (defined $value ? $value : '');
	}
	return @cells;
}

print STDOUT "$filename_in to $filename_out conversion done.\n";
